﻿using System;
using System.IO;
using Microsoft.SPOT.Hardware;
using GTM = Gadgeteer.Modules;
using dotnetwarrior.NetMF.IO;

namespace Gadgeteer.Modules.dotnetwarrior
{
  /// <summary>
  /// A FlashFileSystem module for Microsoft .NET Gadgeteer
  /// </summary>
  public class FlashFileSystem : GTM.Module
  {
    private TinyFileSystem _tfs = null;

    // Note: A constructor summary is auto-generated by the doc builder.
    /// <summary></summary>
    /// <param name="socketNumber">The socket that this module is plugged in to.</param>
    public FlashFileSystem(int socketNumber)
    {
      // This finds the Socket instance from the user-specified socket number.  
      // This will generate user-friendly error messages if the socket is invalid.
      Socket socket = Socket.GetSocket(socketNumber, true, this, null);
      socket.EnsureTypeIsSupported('S', this);

      var spiModule = socket.SPIModule;
      var spiConfig = new SPI.Configuration(socket.CpuPins[6], false, 0, 0, false, true, 12000, spiModule);
      var spi = new SPI(spiConfig);

      var driver = new MX25l3206BlockDriver(spi, socket.CpuPins[5], 4);
      _tfs = new TinyFileSystem(driver);
    }

    /// <summary>
    /// Scan the flash module to verify that it is formatted.
    /// </summary>
    /// <returns>true if formatted else false.</returns>
    public bool CheckIfFormatted()
    {
      return _tfs.CheckIfFormatted();
    }

    /// <summary>
    /// Compacts the file system.
    /// </summary>
    /// <remarks>
    /// Collects all the active clusters and puts them together on the device and formats add the freed sectors.
    /// Frees up all the space occupied by inactive clusters.
    /// </remarks>
    public void Compact()
    {
      _tfs.Compact();
    }

    /// <summary>
    /// Copies a file to a new file.
    /// </summary>
    /// <param name="sourceFileName">The file to copy.</param>
    /// <param name="destFileName">The name of the destination file</param>
    /// <param name="overwrite">Specifies if the destination should be overwritten if it already exists. Default true.</param>
    public void Copy(string sourceFileName, string destFileName, bool overwrite = true)
    {
      _tfs.Copy(sourceFileName, destFileName, overwrite);
    }

    /// <summary>
    /// Creates or overwrites a file.
    /// </summary>
    /// <param name="fileName">Name of the file to create.</param>
    /// <param name="bufferSize">Size of the read/write buffer. 0 for no buffering.</param>
    /// <returns>TinyFileStream that provides stream based access to the file.</returns>
    public Stream Create(string fileName, int bufferSize = 4096)
    {
      return _tfs.Create(fileName, bufferSize);
    }

    /// <summary>
    /// Deletes a file from the device.
    /// </summary>
    /// <param name="fileName">Name of the file to delete.</param>
    /// <remarks>
    /// An IOException will be thrown if the file is open.
    /// </remarks>
    public void Delete(string fileName)
    {
      _tfs.Delete(fileName);
    }

    /// <summary>
    /// Determines if the specified file exists.
    /// </summary>
    /// <param name="fileName">Name of the file to check.</param>
    /// <returns>true if the file exists otherwise false.</returns>
    public bool Exists(string fileName)
    {
      return _tfs.Exists(fileName);
    }

    /// <summary>
    /// Formats the device.
    /// </summary>
    /// <remarks>
    /// This will prepare the device for the file system. 
    /// Note that this will destroy any pre-existing data on the device.
    /// </remarks>
    public void Format()
    {
      _tfs.Format();
    }

    /// <summary>
    /// Gets the creation time of the specified file.
    /// </summary>
    /// <param name="fileName">Name of the file for which to retrieve the creation time.</param>
    /// <returns>The creation time of the specified file.</returns>
    public DateTime GetFileCreationTime(string fileName)
    {
      return _tfs.GetFileCreationTime(fileName);
    }

    /// <summary>
    /// Returns the names of the files in the file system.
    /// </summary>
    /// <returns>An array of the names of the files.</returns>
    public string[] GetFiles()
    {
      return _tfs.GetFiles();
    }

    /// <summary>
    /// Gets the size of the specified file.
    /// </summary>
    /// <param name="fileName">Name of the file for which to retrieve the file size.</param>
    /// <returns>Size of the file in bytes.</returns>
    public long GetFileSize(string fileName)
    {
      return _tfs.GetFileSize(fileName);
    }

    /// <summary>
    /// Get the current statistics of the file system.
    /// </summary>
    /// <returns>Structure with the statistics of the file system.</returns>
    public DeviceStats GetStats()
    {
      return _tfs.GetStats();
    }

    /// <summary>
    /// Mount an existing file system from a device.    
    /// </summary>
    /// <remarks>
    /// Before using the file system it must be mounted. The only exception is
    /// for Format, which will format the device and mount it immediately.
    /// 
    /// The user should ensure that the device has a valid file system.
    /// Currently there are only very basic heuristics to determine if the
    /// device has a valid file system. This will be improved over time.
    /// </remarks>
    public void Mount()
    {
      _tfs.Mount();
    }

    /// <summary>
    /// Moves a file from the source to the destination.
    /// </summary>
    /// <remarks>
    /// Since the Tiny File System does not support directories, move is only used to rename a file.
    /// </remarks>
    /// <param name="sourceFileName">Name of the file to move.</param>
    /// <param name="destFileName">New name of the file.</param>
    public void Move(string sourceFileName, string destFileName)
    {
      _tfs.Move(sourceFileName, destFileName);
    }

    /// <summary>
    /// Opens a TinyFileStream for the specified file.
    /// </summary>
    /// <param name="fileName">Name of the file to open.</param>
    /// <param name="fileMode">Specifies what to do when opening the file.</param>
    /// <param name="bufferSize">Size of the read/write buffer. 0 for no buffering.</param>
    /// <returns>A TinyFileStream which provides stream based access to the file.</returns>
    public Stream Open(string fileName, FileMode fileMode, int bufferSize = 4096)
    {
      return _tfs.Open(fileName, fileMode, bufferSize);
    }

    /// <summary>
    /// Opens a file, reads the content into a byte array and then closes the file.
    /// </summary>
    /// <param name="fileName">Name of the file to read.</param>
    /// <returns>A byte array containing the data from the file.</returns>
    public byte[] ReadAllBytes(string fileName)
    {
      return _tfs.ReadAllBytes(fileName);
    }

    /// <summary>
    /// Creates a new file, writes the byte array to the file and then closes it.
    /// The file is overwritten if it already exists.
    /// </summary>
    /// <param name="fileName">Name of the file to create.</param>
    /// <param name="data">Bytes to be written to the file.</param>
    public void WriteAllBytes(string fileName, byte[] data)
    {
      _tfs.WriteAllBytes(fileName, data);
    }
  }
}
